from pathlib import Path
from typing import List

import allure
import allure_commons.types

from overhave.scenario import OverhaveProjectSettings
from overhave.test_execution import OverhaveAdminLinkSettings


def add_task_links_to_report(project_settings: OverhaveProjectSettings, tasks: List[str]) -> None:
    for task in tasks:
        allure.dynamic.link(
            url=project_settings.get_task_link(task), link_type=allure_commons.types.LinkType.LINK, name=task
        )


def add_git_project_feature_link_to_report(project_settings: OverhaveProjectSettings, filepath: Path) -> None:
    allure.dynamic.link(
        url=project_settings.get_git_feature_url(filepath),
        link_type=allure_commons.types.LinkType.TEST_CASE,
        name=filepath.as_posix(),
    )


def add_admin_feature_link_to_report(admin_link_settings: OverhaveAdminLinkSettings, feature_id: int) -> None:
    allure.dynamic.link(
        url=admin_link_settings.get_feature_url(feature_id),
        link_type=allure_commons.types.LinkType.TEST_CASE,
        name=admin_link_settings.get_feature_link_name(feature_id),
    )



import abc
from typing import Optional, cast

from overhave import db
from overhave.storage.converters import ScenarioModel


class BaseScenarioStorageException(Exception):
    """Base exception for :class:`ScenarioStorage`."""


class ScenarioNotExistsError(BaseScenarioStorageException):
    """Exception for situation without scenario."""


class IScenarioStorage(abc.ABC):
    """Abstract class for feature type storage."""

    @staticmethod
    @abc.abstractmethod
    def get_scenario(scenario_id: int) -> Optional[ScenarioModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_scenario_by_feature_id(feature_id: int) -> ScenarioModel:
        pass

    @staticmethod
    @abc.abstractmethod
    def update_scenario(model: ScenarioModel) -> None:
        pass

    @staticmethod
    @abc.abstractmethod
    def create_scenario(model: ScenarioModel) -> int:
        pass


class ScenarioStorage(IScenarioStorage):
    """Class for feature type storage."""

    @staticmethod
    def get_scenario(scenario_id: int) -> Optional[ScenarioModel]:
        with db.create_session() as session:
            scenario: Optional[db.Scenario] = session.query(db.Scenario).get(scenario_id)
            if scenario is not None:
                return cast(ScenarioModel, ScenarioModel.from_orm(scenario))
            return None

    @staticmethod
    def get_scenario_by_feature_id(feature_id: int) -> ScenarioModel:
        with db.create_session() as session:
            scenario: db.Scenario = session.query(db.Scenario).filter(db.Scenario.feature_id == feature_id).one()
            return cast(ScenarioModel, ScenarioModel.from_orm(scenario))

    @staticmethod
    def update_scenario(model: ScenarioModel) -> None:
        with db.create_session() as session:
            scenario: db.Scenario = session.query(db.Scenario).get(model.id)
            if scenario is None:
                raise ScenarioNotExistsError(f"Scenario with id={model.id} does not exist!")
            scenario.text = model.text

    @staticmethod
    def create_scenario(model: ScenarioModel) -> int:
        with db.create_session() as session:
            scenario = db.Scenario(feature_id=model.feature_id, text=model.text)  # type: ignore
            session.add(scenario)
            session.flush()
            return cast(int, scenario.id)



import abc
from functools import cached_property

from overhave.factory.base_factory import BaseOverhaveFactory, IOverhaveFactory
from overhave.factory.components.abstract_consumer import ITaskConsumerFactory
from overhave.factory.context import OverhavePublicationContext
from overhave.publication import (
    GitlabVersionPublisher,
    IVersionPublisher,
    PublicationManagerType,
    StashVersionPublisher,
    TokenizerClient,
)
from overhave.transport import GitlabHttpClient, PublicationTask, StashHttpClient


class BasePublicationFactoryException(Exception):
    """Base exception for :class:`PublicationFactory`."""


class UrlGitlabTokenizerNotScepifiedIfEnabled(BasePublicationFactoryException):
    """Exception for situation tokenizer is enabled and url env was not set."""


class IPublicationFactory(IOverhaveFactory[OverhavePublicationContext], ITaskConsumerFactory[PublicationTask], abc.ABC):
    """Abstract factory for Overhave publication application."""

    @property
    @abc.abstractmethod
    def publisher(self) -> IVersionPublisher:
        pass


class PublicationFactory(BaseOverhaveFactory[OverhavePublicationContext], IPublicationFactory):
    """Factory for Overhave publication application."""

    context_cls = OverhavePublicationContext

    @cached_property
    def _stash_client(self) -> StashHttpClient:
        return StashHttpClient(self.context.client_settings)  # type: ignore

    @cached_property
    def _gitlab_client(self) -> GitlabHttpClient:
        return GitlabHttpClient(self.context.client_settings)  # type: ignore

    @cached_property
    def _stash_publisher(self) -> StashVersionPublisher:
        return StashVersionPublisher(
            file_settings=self.context.file_settings,
            project_settings=self.context.project_settings,
            feature_storage=self._feature_storage,
            scenario_storage=self._scenario_storage,
            test_run_storage=self._test_run_storage,
            draft_storage=self._draft_storage,
            file_manager=self._file_manager,
            git_initializer=self._git_initializer,
            stash_publisher_settings=self.context.publisher_settings,  # type: ignore
            stash_client=self._stash_client,
        )

    @cached_property
    def _tokenizer_client(self) -> TokenizerClient:
        return TokenizerClient(self.context.tokenizer_client_settings)

    @cached_property
    def _gitlab_publisher(self) -> GitlabVersionPublisher:
        if self._tokenizer_client._settings.enabled and self._tokenizer_client._settings.url is None:
            raise UrlGitlabTokenizerNotScepifiedIfEnabled("Please set correct url for gitlab_tokenizer!")
        return GitlabVersionPublisher(
            file_settings=self.context.file_settings,
            project_settings=self.context.project_settings,
            feature_storage=self._feature_storage,
            scenario_storage=self._scenario_storage,
            test_run_storage=self._test_run_storage,
            draft_storage=self._draft_storage,
            file_manager=self._file_manager,
            git_initializer=self._git_initializer,
            gitlab_publisher_settings=self.context.publisher_settings,  # type: ignore
            gitlab_client=self._gitlab_client,
            tokenizer_client=self._tokenizer_client,
        )

    @property
    def publisher(self) -> IVersionPublisher:
        if self.context.publication_settings.publication_manager_type is PublicationManagerType.GITLAB:
            return self._gitlab_publisher
        return self._stash_publisher

    def process_task(self, task: PublicationTask) -> None:
        return self.publisher.process_publish_task(task)



import abc
import enum
import json
from typing import Any, Dict, TypeVar, Union

from pydantic.main import BaseModel


class RedisStream(str, enum.Enum):
    """Enum that declares Redis streams."""

    TEST = "test"
    PUBLICATION = "publication"
    EMULATION = "emulation"

    @property
    def with_dunder(self) -> str:
        return self.value.replace("-", "_")


class _IRedisTask(BaseModel, abc.ABC):
    @property
    @abc.abstractmethod
    def message(self) -> Dict[bytes, bytes]:
        pass


class BaseRedisTask(_IRedisTask):
    """Base task for Redis streams."""

    data: Any

    @property
    def message(self) -> Dict[bytes, bytes]:
        return {b"data": self.data.json().encode("utf-8")}


class TestRunData(BaseModel):
    """Specific data for test run."""

    __test__ = False

    test_run_id: int


class TestRunTask(BaseRedisTask):
    """Redis stream task for test run."""

    __test__ = False

    data: TestRunData


class PublicationData(BaseModel):
    """Specific data for test run."""

    draft_id: int


class PublicationTask(BaseRedisTask):
    """Redis stream task for test run."""

    data: PublicationData


class EmulationData(BaseModel):
    """Specific data for emulation run."""

    emulation_run_id: int


class EmulationTask(BaseRedisTask):
    """Redis stream task for emulation run."""

    data: EmulationData


TRedisTask = TypeVar("TRedisTask", TestRunTask, EmulationTask, PublicationTask, covariant=True)
AnyRedisTask = Union[TestRunTask, PublicationTask, EmulationTask]


class RedisPendingData(BaseModel):
    """Class that describes pending data from Redis stream."""

    message_id: str
    consumer: str
    time_since_delivered: int
    times_delivered: int


class RedisUnreadData:
    """Class for unread data from Redis stream."""

    def __init__(self, message_id: bytes, message: Dict[bytes, bytes]) -> None:
        self.message_id = message_id.decode()
        self.message = message

    @property
    def decoded_message(self) -> Dict[str, Any]:
        return {key.decode(): json.loads(value.decode("utf-8")) for key, value in self.message.items()}


class RedisContainer(BaseModel):
    """Class for parsing instance of :class:`RedisUnreadData` from Redis stream.

    ```task``` will be parsed to one of declared models.
    """

    task: AnyRedisTask



import enum
import logging
from functools import cached_property
from typing import Optional, cast
from unittest.mock import MagicMock

import allure
from allure_commons._allure import StepContext

from overhave.entities import OverhaveStepContextSettings


class _WrapperPosition(str, enum.Enum):
    ABOVE = "above"
    BELOW = "below"


class StepContextNotDefinedError(Exception):
    """Exception for situation with calling for self._step, when StepContext has not been defined."""


class StepContextRunner:
    """Class for Allure StepContext wrapping during pytest-bdd step execution."""

    def __init__(self, settings: OverhaveStepContextSettings) -> None:
        self._settings = settings
        self._step: Optional[StepContext] = None

    def set_title(self, title: str) -> None:
        self._step = allure.step(title)

    @cached_property
    def _defined_step(self) -> StepContext:
        if isinstance(self._step, StepContext):
            return self._step
        raise StepContextNotDefinedError("Current Allure step not defined!")

    @cached_property
    def _logger(self) -> logging.Logger:
        if self._settings.step_context_logs:
            return logging.getLogger("overhave")
        return cast(logging.Logger, MagicMock())

    def _log_step_state(self, state: str, wrapper_position: _WrapperPosition, symbol: str) -> None:
        description = f"{state.title()} step '{self._defined_step.title}'"
        wrapper = symbol * (len(description) - 2)
        if wrapper_position is _WrapperPosition.ABOVE:
            args = (wrapper, description)
        else:
            args = (description, wrapper)
        self._logger.info("")
        for arg in args:
            self._logger.info("%s", arg)
        self._logger.info("")

    def start(self) -> None:
        self._defined_step.__enter__()
        self._log_step_state(state="started", wrapper_position=_WrapperPosition.ABOVE, symbol="*")

    def stop(self, exception: Optional[BaseException]) -> None:
        if isinstance(exception, BaseException):
            self._log_step_state(state="failed", wrapper_position=_WrapperPosition.BELOW, symbol="!")
            self._defined_step.__exit__(exc_type=type(exception), exc_val=exception, exc_tb=exception.__traceback__)
        else:
            self._log_step_state(state="finished", wrapper_position=_WrapperPosition.BELOW, symbol="-")
            self._defined_step.__exit__(exc_type=None, exc_val=None, exc_tb=None)



from datetime import datetime
from typing import List, Optional

import allure
from pydantic import BaseModel

from overhave.storage import FeatureTypeName


class FeatureInfo(BaseModel):
    """Model for feature info keeping."""

    id: Optional[int]
    name: Optional[str]
    type: Optional[FeatureTypeName]
    tags: Optional[List[str]]
    severity: Optional[allure.severity_level]
    author: Optional[str]
    last_edited_by: Optional[str]
    last_edited_at: Optional[datetime]
    tasks: Optional[List[str]]
    scenarios: Optional[str]


class StrictFeatureInfo(BaseModel):
    """Model for feature info keeping with strict presence of fields."""

    id: int
    name: str
    type: FeatureTypeName
    tags: List[str] = []
    severity: allure.severity_level = allure.severity_level.NORMAL
    author: str
    last_edited_by: str
    last_edited_at: datetime
    tasks: List[str] = []
    scenarios: str



import abc
from functools import cached_property, partial
from multiprocessing.pool import ThreadPool
from typing import Callable, Mapping

from overhave.base_settings import AuthorizationStrategy
from overhave.entities import (
    DefaultAdminAuthorizationManager,
    IAdminAuthorizationManager,
    LDAPAdminAuthorizationManager,
    ReportManager,
    SimpleAdminAuthorizationManager,
)
from overhave.factory.base_factory import IOverhaveFactory
from overhave.factory.components.s3_init_factory import FactoryWithS3ManagerInit
from overhave.factory.context import OverhaveAdminContext
from overhave.storage import IFeatureTypeStorage, SystemUserGroupStorage
from overhave.transport import (
    EmulationTask,
    LDAPAuthenticator,
    PublicationTask,
    RedisProducer,
    RedisStream,
    TestRunTask,
)


class IAdminFactory(IOverhaveFactory[OverhaveAdminContext]):
    """Factory interface for Overhave admin application."""

    @property
    @abc.abstractmethod
    def feature_type_storage(self) -> IFeatureTypeStorage:
        pass

    @cached_property
    @abc.abstractmethod
    def auth_manager(self) -> IAdminAuthorizationManager:
        pass

    @property
    @abc.abstractmethod
    def redis_producer(self) -> RedisProducer:
        pass

    @property
    @abc.abstractmethod
    def report_manager(self) -> ReportManager:
        pass

    @property
    @abc.abstractmethod
    def threadpool(self) -> ThreadPool:
        pass


class AdminFactory(FactoryWithS3ManagerInit[OverhaveAdminContext], IAdminFactory):
    """Factory for Overhave admin application."""

    context_cls = OverhaveAdminContext

    def _get_simple_auth_manager(self) -> partial[SimpleAdminAuthorizationManager]:
        return partial(SimpleAdminAuthorizationManager, system_user_storage=self._system_user_storage)

    def _get_default_auth_manager(self) -> partial[DefaultAdminAuthorizationManager]:
        return partial(DefaultAdminAuthorizationManager, system_user_storage=self._system_user_storage)

    def _get_ldap_auth_manager(self) -> partial[LDAPAdminAuthorizationManager]:
        return partial(
            LDAPAdminAuthorizationManager,
            settings=self.context.ldap_manager_settings,
            system_user_storage=self._system_user_storage,
            system_user_group_storage=SystemUserGroupStorage(),
            ldap_authenticator=LDAPAuthenticator(settings=self.context.ldap_client_settings),
        )

    @property
    def _auth_manager_mapping(
        self,
    ) -> Mapping[AuthorizationStrategy, Callable[[None], partial[IAdminAuthorizationManager]]]:
        return {
            AuthorizationStrategy.SIMPLE: self._get_simple_auth_manager,  # type: ignore
            AuthorizationStrategy.DEFAULT: self._get_default_auth_manager,  # type: ignore
            AuthorizationStrategy.LDAP: self._get_ldap_auth_manager,  # type: ignore
        }

    @cached_property
    def auth_manager(self) -> IAdminAuthorizationManager:
        partial_auth_manager: partial[IAdminAuthorizationManager] = self._auth_manager_mapping[  # type: ignore
            self.context.auth_settings.auth_strategy
        ]()
        return partial_auth_manager()

    @property
    def feature_type_storage(self) -> IFeatureTypeStorage:
        return self._feature_type_storage

    @cached_property
    def _redis_producer(self) -> RedisProducer:
        return RedisProducer(
            settings=self.context.redis_settings,
            mapping={
                TestRunTask: RedisStream.TEST,
                PublicationTask: RedisStream.PUBLICATION,
                EmulationTask: RedisStream.EMULATION,
            },
        )

    @property
    def redis_producer(self) -> RedisProducer:
        return self._redis_producer

    @cached_property
    def _threadpool(self) -> ThreadPool:
        if self.context.admin_settings.consumer_based:
            raise RuntimeError("No threadpool when it's a consumer-based application!")
        return ThreadPool(processes=self.context.admin_settings.threadpool_process_num)

    @property
    def threadpool(self) -> ThreadPool:
        return self._threadpool



from typing import Dict, Optional, Type, Union

from flask_admin.form import JSONField
from flask_login import current_user
from pydantic import BaseModel
from wtforms import Form, ValidationError

from overhave import db
from overhave.admin.views.base import ModelViewConfigured
from overhave.factory import get_admin_factory
from overhave.utils import get_current_time


def _make_dict_from_model(model: Optional[Type[BaseModel]]) -> Optional[Dict[str, Union[int, str]]]:
    if model is not None:
        prepared_dict = {}
        for key, value in model.__fields__.items():
            prepared_dict[key] = value.type_.__name__
        return prepared_dict
    return None


class TestUserView(ModelViewConfigured):
    """View for :class:`TestUser` table."""

    __test__ = False

    create_template = "test_user_create.html"
    edit_template = "test_user_edit.html"

    can_view_details = False
    column_list = ("id", "name", "feature_type", "specification", "allow_update", "created_by", "changed_at")
    column_searchable_list = ("id", "name", "created_by")
    column_filters = ("id", "name", "created_by", "allow_update")
    form_excluded_columns = ("created_at", "emulations", "changed_at")
    form_overrides = dict(specification=JSONField)

    form_extra_fields = {"template": JSONField("Specification format")}
    form_widget_args = {"template": {"readonly": True}}

    column_labels = {
        "name": "Test user name",
        "feature_type": "Feature type",
    }
    column_descriptions = {
        "name": "Custom name for test user",
        "feature_type": "Type of scenarios set, where test user will be used",
        "specification": "Test user specification in JSON format placed below",
        "created_by": "Author of test user set",
        "allow_update": "Property of updating allowance through API",
    }

    _feature_type: Optional[str] = None

    def on_form_prefill(self, form, id) -> None:  # type: ignore  # noqa: A002
        if isinstance(form._obj, db.TestUser):
            self._feature_type = form._obj.feature_type.name

    def get_specification_template(self) -> Optional[Dict[str, Union[int, str]]]:
        factory = get_admin_factory()
        if self._feature_type is None:
            self._feature_type = factory.feature_type_storage.get_default_feature_type().name
        parser = factory.context.project_settings.user_spec_template_mapping.get(self._feature_type)
        return _make_dict_from_model(parser)

    @staticmethod
    def _validate_json(model: db.TestUser) -> None:
        if not isinstance(model.specification, dict):
            raise ValidationError("Could not convert specified data into correct JSON!")
        parser = get_admin_factory().context.project_settings.user_spec_template_mapping.get(model.feature_type.name)
        if parser is not None:
            try:
                parser.parse_obj(model.specification)
            except ValueError:
                raise ValidationError(f"Could not convert specified data into {parser.__name__} model!")

    def on_model_change(self, form: Form, model: db.TestUser, is_created: bool) -> None:
        self._feature_type = model.feature_type.name
        self._validate_json(model)
        if is_created:
            model.created_by = current_user.login
        model.changed_at = get_current_time()

    def on_model_delete(self, model: db.TestUser) -> None:
        if not (current_user.login == model.created_by or current_user.role == db.Role.admin):
            raise ValidationError("Only test user creator or administrator could delete test user!")



import abc
import logging
from typing import Iterable, List, Optional, cast

import sqlalchemy.orm as so
from pydantic.tools import parse_obj_as

from overhave import db
from overhave.storage.converters import FeatureModel
from overhave.utils import get_current_time

logger = logging.getLogger(__name__)


class BaseFeatureStorageException(Exception):
    """Base exception for :class:`FeatureStorage`."""


class FeatureNotExistsError(BaseFeatureStorageException):
    """Error for situation when feature not found."""


class FeatureTagNotExistsError(BaseFeatureStorageException):
    """Error for situation when tag not found."""


class IFeatureStorage(abc.ABC):
    """Abstract class for feature storage."""

    @staticmethod
    @abc.abstractmethod
    def get_feature(feature_id: int) -> Optional[FeatureModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def create_feature(model: FeatureModel) -> int:
        pass

    @staticmethod
    @abc.abstractmethod
    def update_feature(model: FeatureModel) -> None:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_features_by_tag(tag_id: int) -> List[FeatureModel]:
        pass


def _append_tags_to_feature(session: so.Session, feature: db.Feature, tag_ids: Iterable[int]) -> None:
    db_tags: List[db.Tags] = []
    for tag_id in tag_ids:
        db_tag = session.query(db.Tags).get(tag_id)
        if db_tag is None:
            raise FeatureTagNotExistsError(f"Feature tag with id={tag_id} does not exist!")
        logger.info("Append tag with id=%s and value=%s", tag_id, db_tag.value)
        db_tags.append(db_tag)
    feature.feature_tags.extend(db_tags)


class FeatureStorage(IFeatureStorage):
    """Class for feature storage."""

    @staticmethod
    def get_feature(feature_id: int) -> Optional[FeatureModel]:
        with db.create_session() as session:
            feature: Optional[db.Feature] = session.query(db.Feature).get(feature_id)
            if feature is not None:
                return cast(FeatureModel, FeatureModel.from_orm(feature))
            return None

    @staticmethod
    def create_feature(model: FeatureModel) -> int:
        with db.create_session() as session:
            feature = db.Feature(
                name=model.name,
                author=model.author,
                type_id=model.feature_type.id,
                file_path=model.file_path,
                task=model.task,
                severity=model.severity,
                last_edited_at=get_current_time(),
            )
            feature.last_edited_at = model.last_edited_at
            feature.released = model.released
            _append_tags_to_feature(session=session, feature=feature, tag_ids=(tag.id for tag in model.feature_tags))
            session.add(feature)
            session.flush()
            return cast(int, feature.id)

    @staticmethod
    def update_feature(model: FeatureModel) -> None:
        with db.create_session() as session:
            feature: db.Feature = session.query(db.Feature).get(model.id)
            if feature is None:
                raise FeatureNotExistsError(f"Feature with id {model.id} does not exist!")
            feature.name = model.name
            feature.file_path = model.file_path
            feature.task = model.task
            feature.severity = model.severity
            feature.last_edited_by = model.last_edited_by
            feature.last_edited_at = get_current_time()
            feature.released = True
            session.flush()

            existing_tags = {tag.id for tag in feature.feature_tags}
            model_tags = {tag.id for tag in model.feature_tags}
            tags_to_delete = existing_tags.difference(model_tags)
            for tag_id in tags_to_delete:
                db_tag = next(tag for tag in feature.feature_tags if tag.id == tag_id)
                logger.info("Remove tag with id=%s and value=%s", tag_id, db_tag.value)
                feature.feature_tags.remove(db_tag)
            tags_to_append = model_tags.difference(existing_tags)
            _append_tags_to_feature(session=session, feature=feature, tag_ids=tags_to_append)

    @staticmethod
    def get_features_by_tag(tag_id: int) -> List[FeatureModel]:
        with db.create_session() as session:
            feature_ids_query = (
                session.query(db.FeatureTagsAssociationTable)
                .with_entities(db.FeatureTagsAssociationTable.feature_id)
                .filter(db.FeatureTagsAssociationTable.tags_id == tag_id)
                .scalar_subquery()
            )
            features = session.query(db.Feature).filter(db.Feature.id == feature_ids_query).all()
            return parse_obj_as(List[FeatureModel], features)



import logging
import typing
from http import HTTPStatus
from pathlib import Path

import flask
import werkzeug

from overhave import db
from overhave.admin.flask import FlaskLoginManager, get_flask_admin, get_flask_app
from overhave.factory import IAdminFactory, get_publication_factory
from overhave.pytest_plugin import get_proxy_manager
from overhave.storage import UniqueDraftCreationError
from overhave.transport import PublicationData, PublicationTask

logger = logging.getLogger(__name__)

OverhaveAdminApp = typing.NewType("OverhaveAdminApp", flask.Flask)


def _prepare_factory(factory: IAdminFactory) -> None:
    """Resolve necessary settings with :class:`IProxyManager` and prepare :class:`IOverhaveFactory` for usage."""
    proxy_manager = get_proxy_manager()
    proxy_manager.set_factory(factory)
    proxy_manager.patch_pytest()
    proxy_manager.supply_injector_for_collection()
    proxy_manager.injector.collect_configs()


def _resolved_app(factory: IAdminFactory, template_dir: Path) -> flask.Flask:
    """Resolve Flask application with :class:`IOverhaveFactory` and templates directory `template_dir`."""
    flask_app = get_flask_app(template_folder=template_dir.as_posix())
    get_flask_admin(factory).init_app(app=flask_app)
    login_manager = FlaskLoginManager(system_user_storage=factory.system_user_storage, login_view="login")
    login_manager.init_app(flask_app)
    db.ensure_feature_types_exist(factory.feature_extractor.feature_types)  # type: ignore
    return flask_app


def overhave_app(factory: IAdminFactory) -> OverhaveAdminApp:  # noqa: C901
    """Overhave application, based on Flask."""
    current_dir = Path(__file__).parent
    template_dir = current_dir / "templates"
    files_dir = current_dir / "files"

    _prepare_factory(factory)
    flask_app = _resolved_app(factory=factory, template_dir=template_dir)
    flask_app.config["FILES_DIR"] = files_dir

    @flask_app.teardown_request
    def remove_session(exception: typing.Optional[BaseException]) -> None:
        db.current_session.remove()

    @flask_app.route("/reports/<path:request>", methods=["GET", "POST"])
    def get_report(request: str) -> flask.Response:
        if flask.request.method == "POST":
            test_run_id = flask.request.form.get("run_id")
            if test_run_id is None:
                return flask.abort(HTTPStatus.BAD_REQUEST)
            report_precense_resolution = factory.report_manager.get_report_precense_resolution(
                report=request, run_id=int(test_run_id)
            )
            if not report_precense_resolution.exists:
                if report_precense_resolution.not_ready:
                    return typing.cast(
                        flask.Response, flask.redirect(f"/reports/{request}", code=HTTPStatus.TEMPORARY_REDIRECT)
                    )
                return flask.abort(HTTPStatus.NOT_FOUND)
        return flask.send_from_directory(factory.context.file_settings.tmp_reports_dir, request)

    @flask_app.route("/emulations/<path:url>")
    def go_to_emulation(url: str) -> werkzeug.Response:
        return flask.redirect(factory.context.emulation_settings.get_emulation_url(url))

    @flask_app.route("/pull_request/<int:run_id>")
    def publish_feature(run_id: int) -> werkzeug.Response:
        published_by = flask.request.args.get("published_by")
        if not isinstance(published_by, str):
            flask.flash("Parameter 'published_by' should be specified for version's creation!", category="error")
            return flask.redirect(flask.url_for("testrun.details_view", id=run_id))
        try:
            draft_id = factory.draft_storage.save_draft(
                test_run_id=run_id, published_by=published_by, status=db.DraftStatus.REQUESTED
            )
        except UniqueDraftCreationError:
            logger.exception("Error while creation draft!")
            flask.flash(
                "Requested publication contains scenario which identical to the previous version!", category="warning"
            )
            return flask.redirect(flask.url_for("testrun.details_view", id=run_id))
        if not factory.context.admin_settings.consumer_based:
            factory.threadpool.apply_async(get_publication_factory().publisher.publish_version, args=(draft_id,))
        if factory.context.admin_settings.consumer_based and not factory.redis_producer.add_task(
            PublicationTask(data=PublicationData(draft_id=draft_id))
        ):
            flask.flash("Problems with Redis service! TestRunTask has not been sent.", category="error")
            return flask.redirect(flask.url_for("testrun.details_view", id=run_id))
        return flask.redirect(flask.url_for("draft.details_view", id=draft_id))

    @flask_app.route("/files/<path:file>")
    def get_files(file: str) -> flask.Response:
        return flask.send_from_directory(files_dir, file)

    @flask_app.route("/favicon.ico")
    def favicon() -> flask.Response:
        return flask.send_from_directory(files_dir, "favicon.ico", mimetype="image/vnd.microsoft.icon")

    return OverhaveAdminApp(flask_app)



import abc

from overhave.db import DraftStatus
from overhave.entities import OverhaveFileSettings
from overhave.publication.abstract_publisher import IVersionPublisher
from overhave.scenario import FileManager, OverhaveProjectSettings, generate_task_info
from overhave.storage import IDraftStorage, IFeatureStorage, IScenarioStorage, ITestRunStorage, PublisherContext
from overhave.transport import PublicationTask


class BaseVersionPublisherException(Exception):
    """Base exception for :class:`BaseVersionPublisher`."""


class FeatureNotExistsError(BaseVersionPublisherException):
    """Exception for situation with not existing Feature."""


class DraftNotExistsError(BaseVersionPublisherException):
    """Exception for situation with not existing Draft."""


class TestRunNotExistsError(BaseVersionPublisherException):
    """Exception for situation with not existing TestRun."""


class ScenarioNotExistsError(BaseVersionPublisherException):
    """Exception for situation with not existing Scenario."""


class NullablePullRequestUrlError(BaseVersionPublisherException):
    """Exception for nullable merge-request in selected Draft."""


class BaseVersionPublisher(IVersionPublisher, abc.ABC):
    """Class for feature version's pull requests management relative to Atlassian Stash API."""

    def __init__(
        self,
        file_settings: OverhaveFileSettings,
        project_settings: OverhaveProjectSettings,
        feature_storage: IFeatureStorage,
        scenario_storage: IScenarioStorage,
        test_run_storage: ITestRunStorage,
        draft_storage: IDraftStorage,
        file_manager: FileManager,
    ) -> None:
        self._file_settings = file_settings
        self._project_settings = project_settings
        self._feature_storage = feature_storage
        self._scenario_storage = scenario_storage
        self._test_run_storage = test_run_storage
        self._draft_storage = draft_storage
        self._file_manager = file_manager

    def process_publish_task(self, task: PublicationTask) -> None:
        self.publish_version(task.data.draft_id)

    def _compile_context(self, draft_id: int) -> PublisherContext:
        draft = self._draft_storage.get_draft(draft_id)
        if not draft:
            raise DraftNotExistsError(f"Draft with id={draft_id} does not exist!")
        test_run = self._test_run_storage.get_test_run(draft.test_run_id)
        if not test_run:
            raise TestRunNotExistsError(f"TestRun with id={draft.test_run_id} does not exist!")
        feature = self._feature_storage.get_feature(draft.feature_id)
        if not feature:
            raise FeatureNotExistsError(f"Feature with id={draft.feature_id} does not exist!")
        scenario = self._scenario_storage.get_scenario(test_run.scenario_id)
        if not scenario:
            raise ScenarioNotExistsError(f"Scenario with id={test_run.scenario_id} does not exist!")
        return PublisherContext(
            feature=feature,
            scenario=scenario,
            test_run=test_run,
            draft=draft,
            target_branch=f"bdd-feature-{feature.id}",
        )

    def _compile_publication_description(self, context: PublisherContext) -> str:
        return "\n".join(
            (
                f"Feature ID: {context.feature.id}. Type: '{context.feature.feature_type.name}'.",
                f"Created by: @{context.feature.author} at {context.feature.created_at.strftime('%d-%m-%Y %H:%M:%S')}.",
                f"Last edited by: @{context.feature.last_edited_by}.",
                f"PR from Test Run ID: {context.test_run.id}. Executed by: @{context.test_run.executed_by}",
                f"Published by: @{context.draft.published_by}.",
                generate_task_info(tasks=context.feature.task, header=self._project_settings.tasks_keyword),
            )
        )

    def _save_as_duplicate(self, context: PublisherContext) -> None:
        previous_draft = self._draft_storage.get_previous_feature_draft(context.feature.id)
        if previous_draft.pr_url is None:
            raise NullablePullRequestUrlError(
                "Previous draft with id=%s has not got pull-request URL!", previous_draft.id
            )
        if previous_draft.published_at is None:
            raise RuntimeError(f"Somebody has published this scenario at {previous_draft.published_at}")

        self._draft_storage.save_response(
            draft_id=context.draft.id,
            pr_url=context.draft.pr_url or previous_draft.pr_url,
            published_at=previous_draft.published_at,
            status=DraftStatus.DUPLICATE,
            traceback=context.draft.traceback,
        )



import logging
from typing import List

import fastapi

from overhave.api.deps import get_emulation_storage
from overhave.storage import EmulationRunModel, IEmulationStorage

logger = logging.getLogger(__name__)


def emulation_run_list_handler(
    test_user_id: int,
    emulation_storage: IEmulationStorage = fastapi.Depends(get_emulation_storage),
) -> List[EmulationRunModel]:
    logger.info("Getting %s list with test_user_id=%s...", EmulationRunModel.__name__, test_user_id)
    return emulation_storage.get_emulation_runs_by_test_user_id(test_user_id=test_user_id)



import abc


class IOverhaveSynchronizer(abc.ABC):
    """Abstract class for synchronization between git and database."""

    @abc.abstractmethod
    def synchronize(self, create_db_features: bool = False, pull_repository: bool = False) -> None:
        pass



import logging
from types import TracebackType
from typing import Iterator, List, Sequence, Type

import walrus

from overhave.entities.settings import OverhaveRedisSettings
from overhave.transport.redis.objects import RedisPendingData, RedisStream, RedisUnreadData
from overhave.transport.redis.template import RedisTemplate

logger = logging.getLogger(__name__)


class RedisConsumer(RedisTemplate):
    """Class for consuming tasks from Redis stream ```stream_name```."""

    def __init__(self, settings: OverhaveRedisSettings, stream_name: RedisStream):
        super().__init__(settings)
        self._stream_name = stream_name

    @property
    def _consumer_group(self) -> walrus.ConsumerGroup:
        consumer_group = self._database.consumer_group(f"cg-{self._stream_name}", (self._stream_name,))
        consumer_group.create()
        return consumer_group

    @property
    def _stream(self) -> walrus.containers.ConsumerGroupStream:
        return getattr(self._consumer_group, self._stream_name.with_dunder)

    def _clean_pending(self) -> None:
        pending_messages = self._stream.pending()
        models: List[RedisPendingData] = [RedisPendingData.parse_obj(msg) for msg in pending_messages]
        if models:
            message_ids = [x.message_id for x in models]
            self._stream.claim(*message_ids)
            self._stream.ack(*message_ids)
            logger.info("Clean all pending messages for stream %s: %s", self._stream_name, models)

    def _consume(self) -> Sequence[RedisUnreadData]:
        messages = self._stream.read(count=self._settings.redis_read_count, block=self._settings.timeout_milliseconds)
        objects: List[RedisUnreadData] = []
        for msg in messages:
            data = RedisUnreadData(*msg)
            logger.info("Message from redis: %s", data)
            self._stream.ack(data.message_id)
            objects.append(data)
        return objects

    def __enter__(self) -> None:
        logger.info("Starting consuming from %s...", self._stream_name)
        self._clean_pending()

    def __iter__(self) -> Iterator[Sequence[RedisUnreadData]]:
        while True:
            try:
                logger.debug("Check messages...")
                messages = self._consume()
                if messages:
                    logger.debug("Has messages, return them")
                    yield messages
                continue
            except Exception:
                logger.exception("Error while trying to consume message from redis!")
            raise StopIteration()

    def __exit__(self, exc_type: Type[Exception], exc_val: Exception, exc_tb: TracebackType) -> None:
        pass



from overhave.base_settings import BaseOverhavePrefix


class OverhaveScenarioCompilerSettings(BaseOverhavePrefix):
    """Settings for scenario compiling and parsing."""

    tag_prefix: str = "@"
    id_prefix: str = "ID:"
    created_by_prefix: str = "# created by"
    last_edited_by_prefix: str = "last edited by"
    published_by_prefix: str = "published by"
    severity_prefix: str = "@severity."

    time_delimiter: str = ","
    time_format: str = "%d-%m-%Y %H:%M:%S"
    blocks_delimiter: str = "|"

    @property
    def severity_keyword(self) -> str:
        return self.severity_prefix.removeprefix(self.tag_prefix)



import logging
from typing import List

import fastapi

from overhave.api.deps import get_feature_type_storage
from overhave.storage import FeatureTypeModel, IFeatureTypeStorage

logger = logging.getLogger(__name__)


def feature_types_list_handler(
    feature_type_storage: IFeatureTypeStorage = fastapi.Depends(get_feature_type_storage),
) -> List[FeatureTypeModel]:
    logger.info("Getting %s list...", FeatureTypeModel.__name__)
    return feature_type_storage.get_all_feature_types()



# flake8: noqa
from .models import FeatureInfo, StrictFeatureInfo
from .parser import FeatureNameParsingError, NullableFeatureIdError, ScenarioParser, StrictFeatureParsingError
from .settings import OverhaveScenarioParserSettings



from __future__ import annotations

import datetime
from typing import List, Optional, cast

import allure
import sqlalchemy as sa
import sqlalchemy_utils as su
from flask import url_for
from sqlalchemy import orm as so

from overhave.db.base import BaseTable, PrimaryKeyMixin, PrimaryKeyWithoutDateMixin
from overhave.db.statuses import DraftStatus, EmulationStatus, TestReportStatus, TestRunStatus
from overhave.db.users import UserRole


class FeatureType(BaseTable, PrimaryKeyWithoutDateMixin):
    """Feature types table."""

    name = sa.Column(sa.Text, unique=True, nullable=False, doc="Feature types choice")

    def __repr__(self) -> str:
        return cast(str, self.name.upper())


class Tags(BaseTable, PrimaryKeyMixin):
    """Feature tags table."""

    value = sa.Column(sa.String(), nullable=False, doc="Feature tags choice", unique=True)
    created_by = sa.Column(sa.String(), sa.ForeignKey(UserRole.login), doc="Author login", nullable=False)

    def __repr__(self) -> str:
        return cast(str, self.value)

    def __init__(self, value: str, created_by: str) -> None:
        self.value = value
        self.created_by = created_by


@su.generic_repr("id", "name", "last_edited_by")
class Feature(BaseTable, PrimaryKeyMixin):
    """Features table."""

    name = sa.Column(sa.String(), doc="Feature name", nullable=False, unique=True)
    author = sa.Column(
        sa.String(), sa.ForeignKey(UserRole.login), doc="Feature author login", nullable=False, index=True
    )
    type_id = sa.Column(sa.Integer(), sa.ForeignKey(FeatureType.id), nullable=False, doc="Feature types choice")
    file_path = sa.Column(sa.String(), doc="Feature file path", nullable=False, unique=True)
    task = sa.Column(sa.ARRAY(sa.String()), doc="Feature tasks list", nullable=False)
    last_edited_by = sa.Column(sa.String(), doc="Last feature editor login", nullable=False)
    last_edited_at = sa.Column(sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now())
    released = sa.Column(sa.Boolean, doc="Feature release state boolean", nullable=False, default=False)
    severity = sa.Column(
        sa.Enum(allure.severity_level),
        nullable=False,
        default=allure.severity_level.NORMAL,
        doc="Feature severity choice",
    )

    feature_type = so.relationship(FeatureType)
    feature_tags = so.relationship(Tags, order_by=Tags.value, secondary="feature_tags_association_table")

    def __init__(
        self,
        name: str,
        author: str,
        type_id: int,
        file_path: str,
        task: List[str],
        severity: allure.severity_level,
        last_edited_at: datetime.datetime,
    ) -> None:
        self.name = name
        self.author = author
        self.type_id = type_id
        self.file_path = file_path
        self.task = task
        self.last_edited_by = author
        self.severity = severity
        self.last_edited_at = last_edited_at


class FeatureTagsAssociationTable(BaseTable, PrimaryKeyWithoutDateMixin):
    """Association table between features and tags."""

    extend_existing = True

    tags_id = sa.Column(sa.Integer(), sa.ForeignKey(Tags.id))
    feature_id = sa.Column(sa.Integer(), sa.ForeignKey(Feature.id))


@su.generic_repr("feature_id")
class Scenario(BaseTable, PrimaryKeyMixin):
    """Scenarios table."""

    feature_id = sa.Column(sa.Integer(), sa.ForeignKey(Feature.id), nullable=False, unique=True)
    text = sa.Column(sa.Text(), doc="Text storage for scenarios in feature")

    feature = so.relationship(Feature, backref=so.backref("scenario", cascade="all, delete-orphan"))


class TestRun(BaseTable, PrimaryKeyMixin):
    """Test runs table."""

    __test__ = False

    scenario_id = sa.Column(sa.Integer(), sa.ForeignKey(Scenario.id), nullable=False, index=True)
    name = sa.Column(sa.String(), nullable=False)
    start = sa.Column(sa.DateTime(timezone=True), doc="Test start time")
    end = sa.Column(sa.DateTime(timezone=True), doc="Test finish time")
    status = sa.Column(sa.Enum(TestRunStatus), doc="Current test status", nullable=False)
    report_status = sa.Column(sa.Enum(TestReportStatus), doc="Report generation result", nullable=False)
    executed_by = sa.Column(sa.String(), sa.ForeignKey(UserRole.login), doc="Test executor login", nullable=False)
    report = sa.Column(sa.String(), doc="Relative report URL")
    traceback = sa.Column(sa.Text(), doc="Text storage for error traceback")

    scenario = so.relationship(Scenario, backref=so.backref("test_runs", cascade="all, delete-orphan"))


class DraftQuery(so.Query):
    """Scenario versions table."""

    def as_unique(self, test_run_id: int, published_by: str, status: DraftStatus) -> Draft:
        with self.session.no_autoflush:
            run = self.session.query(TestRun).get(test_run_id)
            if run is None:
                raise RuntimeError(f"Unknown test_run_id={test_run_id}!")
            draft: Optional[Draft] = (
                self.session.query(Draft)
                .filter(Draft.test_run_id == run.id, Draft.text == run.scenario.text)
                .one_or_none()
            )
        if draft:
            return draft
        return Draft(
            feature_id=run.scenario.feature_id,
            test_run_id=test_run_id,
            text=run.scenario.text,
            published_by=published_by,
            status=status,
        )


class Draft(BaseTable, PrimaryKeyMixin):
    """Scenario versions table."""

    __query_cls__ = DraftQuery

    feature_id = sa.Column(sa.Integer(), sa.ForeignKey(Feature.id), nullable=False, index=True)
    test_run_id = sa.Column(sa.Integer(), sa.ForeignKey(TestRun.id), nullable=False)
    text = sa.Column(sa.Text(), doc="Released scenario text")
    pr_url = sa.Column(sa.String(), doc="Absolute pull-request URL", nullable=True)
    published_by = sa.Column(sa.String(), sa.ForeignKey(UserRole.login), doc="Draft publisher login", nullable=False)
    published_at = sa.Column(sa.DateTime(timezone=True), doc="Publication time")
    traceback = sa.Column(sa.Text(), doc="Text storage for error traceback", nullable=True)
    status = sa.Column(sa.Enum(DraftStatus), doc="Version publishing status", nullable=False)

    feature = so.relationship(Feature, backref=so.backref("versions", cascade="all, delete-orphan"))

    __table_args__ = (sa.UniqueConstraint(test_run_id),)

    def __html__(self) -> str:
        return f'<a href="{url_for("draft.details_view", id=self.id)}">Draft: {self.id}</a>'

    def __init__(self, feature_id: int, test_run_id: int, text: str, published_by: str, status: DraftStatus) -> None:
        self.feature_id = feature_id
        self.test_run_id = test_run_id
        self.text = text
        self.published_by = published_by
        self.status = status


@su.generic_repr("id", "name", "created_by")
class TestUser(BaseTable, PrimaryKeyMixin):
    """Test users table."""

    __test__ = False

    name = sa.Column(sa.String(), nullable=False, unique=True)
    feature_type_id = sa.Column(sa.Integer(), sa.ForeignKey(FeatureType.id), nullable=False, doc="Feature types choice")
    specification = sa.Column(sa.JSON(none_as_null=True))
    created_by = sa.Column(sa.String(), sa.ForeignKey(UserRole.login), doc="Author login", nullable=False)
    allow_update = sa.Column(sa.Boolean(), doc="User updating allowance", nullable=False, default=False)
    changed_at = sa.Column(sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now())

    feature_type = so.relationship(FeatureType)


class Emulation(BaseTable, PrimaryKeyMixin):
    """Emulation templates table."""

    name = sa.Column(sa.String(), nullable=False, unique=True)
    test_user_id = sa.Column(sa.Integer(), sa.ForeignKey(TestUser.id), nullable=False, doc="Test user ID")
    command = sa.Column(sa.String(), nullable=False, doc="Command for emulator's execution")
    created_by = sa.Column(sa.String(), sa.ForeignKey(UserRole.login), doc="Author login", nullable=False)
    test_user = so.relationship(TestUser, backref=so.backref("emulations", cascade="all, delete-orphan"))


class EmulationRun(BaseTable, PrimaryKeyMixin):
    """Emulation runs table."""

    __tablename__ = "emulation_run"  # type: ignore
    emulation_id = sa.Column(sa.Integer(), sa.ForeignKey(Emulation.id), nullable=False, index=True)
    status = sa.Column(sa.Enum(EmulationStatus), doc="Current emulation status", nullable=False)
    changed_at = sa.Column(sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now())
    port = sa.Column(sa.Integer(), doc="Port for emulation")
    traceback = sa.Column(sa.Text(), doc="Text storage for error traceback")
    initiated_by = sa.Column(
        sa.String(), sa.ForeignKey(UserRole.login), doc="Initiator of start emulation", nullable=False
    )

    emulation = so.relationship(Emulation, backref=so.backref("emulation_runs", cascade="all, delete-orphan"))

    def __init__(self, emulation_id: int, initiated_by: str) -> None:
        self.emulation_id = emulation_id
        self.status = EmulationStatus.CREATED
        self.initiated_by = initiated_by



import logging
from datetime import timedelta
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

import boto3
import botocore.exceptions
import urllib3
from boto3_type_annotations.s3 import Client
from pydantic.tools import parse_obj_as
from yarl import URL

from overhave.transport.s3.models import BucketModel, DeletionResultModel, ObjectModel
from overhave.transport.s3.objects import OverhaveS3Bucket
from overhave.transport.s3.settings import OverhaveS3ManagerSettings
from overhave.utils import make_url

logger = logging.getLogger(__name__)


class BaseS3ManagerException(Exception):
    """Base exception for :class:`S3Manager`."""


class UndefinedClientException(BaseS3ManagerException):
    """Exception for situation with not initialized client in :class:`S3Manager`."""


class InvalidEndpointError(BaseS3ManagerException):
    """Exception for ValueError with invalid endpoint from boto3."""


class InvalidCredentialsError(BaseS3ManagerException):
    """Exception for situation with invalid credentials from boto3."""


class EndpointConnectionError(BaseS3ManagerException):
    """Exception for situation with endpoint connection error from boto3."""


class ClientError(BaseS3ManagerException):
    """Exception for situation with client error from boto3."""


class EmptyObjectsListError(BaseS3ManagerException):
    """Exception for situation with empty object list."""


def _s3_error(msg: str):  # type: ignore
    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        def wrapper(*args: Any, **kwargs: Dict[str, Any]) -> Any:
            try:
                return func(*args, **kwargs)
            except botocore.exceptions.ClientError as e:
                raise ClientError(msg) from e

        return wrapper

    return decorator


class S3Manager:
    """Class for s3 management with boto3 client."""

    def __init__(self, settings: OverhaveS3ManagerSettings):
        self._settings = settings
        self._client: Optional[Client] = None

    def initialize(self) -> None:
        if not self._settings.enabled:
            logger.info("S3Manager disabled and has not been initialized.")
            return
        self._client = self._get_client(self._settings)
        if self._settings.autocreate_buckets:
            self._ensure_buckets_exists()

    @property
    def enabled(self) -> bool:
        return self._client is not None

    @staticmethod
    def _get_client(settings: OverhaveS3ManagerSettings) -> Client:
        if not settings.verify:
            logger.warning("Verification disabled in '%s', so ignore 'urllib3' warnings.", type(settings).__name__)
            urllib3.disable_warnings()
        try:
            client = boto3.client(
                "s3",
                region_name=settings.region_name,
                verify=settings.verify,
                endpoint_url=settings.url,
                aws_access_key_id=settings.access_key,
                aws_secret_access_key=settings.secret_key,
            )
            logger.info("s3 client successfully initialized.")
            return client  # noqa: R504
        except ValueError as e:
            raise InvalidEndpointError("Incorrect specified URL for s3 client!") from e
        except botocore.exceptions.ClientError as e:
            raise InvalidCredentialsError("Invalid credentials for s3 client!") from e
        except botocore.exceptions.EndpointConnectionError as e:
            raise EndpointConnectionError(f"Could not connect to '{settings.url}'!") from e

    @property
    def _ensured_client(self) -> Client:
        if self._client is None:
            raise UndefinedClientException("s3 client has not been initialized!")
        return self._client

    def _ensure_buckets_exists(self) -> None:
        remote_buckets = self._get_buckets()
        logger.info("Existing remote s3 buckets: %s", remote_buckets)
        bucket_names = [model.name for model in remote_buckets]
        for bucket in list(filter(lambda x: x.value not in bucket_names, OverhaveS3Bucket)):
            self.create_bucket(bucket.value)
        logger.info("Successfully ensured existence of Overhave service buckets.")

    @_s3_error(msg="Error while getting buckets list!")
    def _get_buckets(self) -> List[BucketModel]:
        response = self._ensured_client.list_buckets()
        return parse_obj_as(List[BucketModel], response.get("Buckets"))

    @_s3_error(msg="Error while creating bucket!")
    def create_bucket(self, bucket: str) -> None:
        logger.info("Creating bucket %s...", bucket)
        kwargs: Dict[str, Any] = {"Bucket": bucket}
        if isinstance(self._settings.region_name, str):
            kwargs["CreateBucketConfiguration"] = {"LocationConstraint": self._settings.region_name}
        self._ensured_client.create_bucket(**kwargs)
        logger.info("Bucket %s successfully created.", bucket)

    def upload_file(self, file: Path, bucket: OverhaveS3Bucket) -> bool:
        logger.info("Start uploading file '%s'...", file.name)
        try:
            self._ensured_client.upload_file(file.as_posix(), bucket.value, file.name)
            logger.info("File '%s' successfully uploaded", file.name)
            return True
        except botocore.exceptions.ClientError:
            logger.exception("Could not upload file to s3 cloud!")
            return False

    @_s3_error(msg="Error while getting bucket objects list!")
    def get_bucket_objects(self, bucket: str) -> List[ObjectModel]:
        response = self._ensured_client.list_objects(Bucket=bucket)
        logger.debug("List objects response:\n%s", response)
        return parse_obj_as(List[ObjectModel], response.get("Contents"))

    @_s3_error(msg="Error while deleting bucket objects!")
    def delete_bucket_objects(self, bucket: str, objects: List[ObjectModel]) -> DeletionResultModel:
        if not objects:
            raise EmptyObjectsListError("No one object specified for deletion!")
        logger.info("Deleting items %s...", [obj.name for obj in objects])
        response = self._ensured_client.delete_objects(
            Bucket=bucket,
            Delete={"Objects": [{"Key": obj.name} for obj in objects]},
        )
        logger.debug("Delete objects response:\n%s", response)
        return DeletionResultModel.parse_obj(response)

    def _ensure_bucket_clean(self, bucket: str) -> None:
        objects = self.get_bucket_objects(bucket)
        if not objects:
            logger.info("Has not got any objects in bucket '%s'.", bucket)
            return
        deletion_result = self.delete_bucket_objects(bucket=bucket, objects=objects)
        if len(deletion_result.deleted) != len(objects):
            logger.warning("Expected %s deleted objects, got %s!", len(objects), len(deletion_result.deleted))
            logger.warning("Errors while deleted items:\n%s", deletion_result.errors)
        logger.info("Items %s successfully deleted.", [obj.name for obj in deletion_result.deleted])

    @_s3_error("Error while deleting bucket!")
    def delete_bucket(self, bucket: str, force: bool = False) -> None:
        if force:
            self._ensure_bucket_clean(bucket)
        logger.info("Deleting bucket '%s'...", bucket)
        self._ensured_client.delete_bucket(Bucket=bucket)
        logger.info("Bucket '%s' successfully deleted.", bucket)

    def download_file(self, filename: str, dir_to_save: Path, bucket: str) -> bool:
        logger.info("Start downloading file '%s'...", filename)
        try:
            self._ensured_client.download_file(
                Bucket=bucket, Key=filename, Filename=(dir_to_save / filename).as_posix()
            )
            logger.info("File '%s' successfully downloaded", filename)
            return True
        except botocore.exceptions.ClientError:
            logger.exception("Could not download file from s3 cloud!")
            return False

    def create_presigned_url(self, bucket: str, object_name: str, expiration: timedelta) -> Optional[URL]:
        try:
            response = self._ensured_client.generate_presigned_url(
                "get_object", Params={"Bucket": bucket, "Key": object_name}, ExpiresIn=int(expiration.total_seconds())
            )
            logger.info("Created presigned URL: '%s'", response)
            return make_url(response)
        except botocore.exceptions.ClientError:
            logging.exception("Could not create presigned URL for s3 cloud object '%s'!", object_name)
            return None



import logging
from contextlib import contextmanager
from typing import Any, Iterator, List

import sqlalchemy.orm as so
from sqlalchemy.exc import ProgrammingError

from overhave.db.base import Session
from overhave.db.tables import FeatureType

logger = logging.getLogger(__name__)


@contextmanager
def create_session(**kwargs: Any) -> Iterator[so.Session]:
    """Provide a transactional scope around a series of operations."""
    new_session = Session(**kwargs)
    try:
        yield new_session
        new_session.commit()
    except Exception:
        new_session.rollback()
        raise
    finally:
        new_session.close()


def ensure_feature_types_exist(feature_types: List[str]) -> None:
    with create_session() as session:
        try:
            for feature_type in feature_types:
                existing_type = session.query(FeatureType).filter_by(name=feature_type).one_or_none()
                if existing_type is not None:
                    continue
                session.add(FeatureType(name=feature_type))  # type: ignore
                logger.info("Created feature type '%s'", feature_type)
        except ProgrammingError:
            logger.exception("Could not create feature types dynamically!")



from functools import cache

from overhave.pytest_plugin.helpers import OverhaveTagController
from overhave.pytest_plugin.helpers.allure_utils.description_manager import DescriptionManager
from overhave.pytest_plugin.helpers.allure_utils.step_context_runner import StepContextRunner


@cache
def get_step_context_runner() -> StepContextRunner:
    from overhave.factory import get_test_execution_factory

    return StepContextRunner(settings=get_test_execution_factory().context.step_context_settings)


@cache
def get_description_manager() -> DescriptionManager:
    from overhave.factory import get_test_execution_factory

    return DescriptionManager(settings=get_test_execution_factory().context.description_manager_settings)


@cache
def get_tag_controller() -> OverhaveTagController:
    return OverhaveTagController()



import logging
from pprint import pformat
from typing import List, Optional

from pydantic import SecretStr

from overhave.db import Role
from overhave.entities.auth_managers.base import BaseAdminAuthorizationManager
from overhave.entities.auth_managers.ldap.settings import OverhaveLdapManagerSettings
from overhave.storage import ISystemUserGroupStorage, ISystemUserStorage, SystemUserModel
from overhave.transport.ldap.authenticator import LDAPAuthenticator

logger = logging.getLogger(__name__)


class LDAPEmptyAdminGroupError(ValueError):
    """Exception for situation with empty LDAP admin group."""


class LDAPAdminAuthorizationManager(BaseAdminAuthorizationManager):
    """Class for user auth_managers via LDAP.

    Manager authorize users using remote LDAP server. Each user should use his LDAP credentials.
    LDAP server returns user groups. If user in default 'admin' group or his groups list contains admin group - user
    will be authorized. If user already placed in database - user will be authorized too. No one password stores.
    """

    def __init__(
        self,
        settings: OverhaveLdapManagerSettings,
        system_user_storage: ISystemUserStorage,
        system_user_group_storage: ISystemUserGroupStorage,
        ldap_authenticator: LDAPAuthenticator,
    ):
        super().__init__(system_user_storage)
        self._settings = settings
        self._system_user_group_storage = system_user_group_storage
        self._ldap_authenticator = ldap_authenticator

    def _reassign_role_if_neccessary(self, user: SystemUserModel, user_groups: List[str]) -> None:
        if self._settings.ldap_admin_group not in user_groups:
            return
        user.role = Role.admin
        self._system_user_storage.update_user_role(user_model=user)

    def authorize_user(self, username: str, password: SecretStr) -> Optional[SystemUserModel]:
        logger.debug("Try to authorize user '%s'...", username)
        user_groups = self._ldap_authenticator.get_user_groups(username, password)
        if not user_groups:
            logger.info("LDAP user '%s' does not exist!", username)
            return None
        logger.debug("LDAP user groups: \n %s", pformat(user_groups))
        user = self._system_user_storage.get_user_by_credits(login=username)
        if user is not None:
            self._reassign_role_if_neccessary(user=user, user_groups=user_groups)
            return user
        logger.debug("Have not found user with username '%s'!", username)
        if self._system_user_group_storage.has_any_group(user_groups) or self._settings.ldap_admin_group in user_groups:
            user = self._system_user_storage.create_user(login=username)
            self._reassign_role_if_neccessary(user=user, user_groups=user_groups)
            return user
        logger.debug("Received user groups (%s) are not supplied with database groups!", user_groups)
        return None



# flake8: noqa
from .deps import get_description_manager, get_step_context_runner
from .helpers import (
    DescriptionManager,
    StepContextNotDefinedError,
    StepContextRunner,
    get_feature_info_from_item,
    get_scenario,
)
from .plugin_resolver import IPluginResolver, PluginResolver
from .proxy_manager import IProxyManager, ProxyManager, get_proxy_manager



from overhave.base_settings import BaseOverhavePrefix


class OverhaveScenarioParserSettings(BaseOverhavePrefix):
    """Settings for scenario parser."""

    parser_strict_mode: bool = False  # ScenarioParser mode; if strict - necessary fields will be validated



import logging
from pathlib import Path
from typing import List

from overhave.entities import BaseFileExtractor, GitRepositoryInitializer, OverhaveFileSettings
from overhave.scenario import StrictFeatureParsingError
from overhave.scenario.parser import NullableFeatureIdError, ScenarioParser

logger = logging.getLogger(__name__)


class BaseFeatureValidatorException(Exception):
    """Base exception for :class:`FeatureValidator`."""


class FeaturesWithoutIDPresenceError(BaseFeatureValidatorException):
    """Exception for situation with nullable ID features presence."""


class IncorrectFeaturesPresenceError(BaseFeatureValidatorException):
    """Exception for situation with incorrect features presence."""


class FeatureValidator(BaseFileExtractor):
    """Class for features validation."""

    def __init__(
        self,
        file_settings: OverhaveFileSettings,
        scenario_parser: ScenarioParser,
        git_initializer: GitRepositoryInitializer,
    ):
        super().__init__(extenstion=file_settings.feature_suffix)
        self._file_settings = file_settings
        self._scenario_parser = scenario_parser
        self._git_initializer = git_initializer

        self._scenario_parser.set_strict_mode(True)
        self._nullable_id_features: List[Path] = []
        self._incorrect_features: List[Path] = []

    def validate(self, raise_if_nullable_id: bool = False, pull_repository: bool = False) -> None:
        if pull_repository:
            self._git_initializer.pull()
        logger.info("Start validation...")
        all_features = self._extract_recursively(self._file_settings.features_dir)
        for feature_file in all_features:
            logger.info("Read feature from file %s...", feature_file.as_posix())
            try:
                feature_info = self._scenario_parser.parse(feature_file.read_text())
            except NullableFeatureIdError:
                logger.warning("Feature has not got suitable Overhave ID!")
                self._nullable_id_features.append(feature_file)
            except StrictFeatureParsingError:
                logger.exception("Feature has incorrect format!")
                self._incorrect_features.append(feature_file)
            else:
                logger.info("Feature successfully parsed: %s", feature_info)
        if self._incorrect_features:
            raise IncorrectFeaturesPresenceError(
                f"Features with incorrect format: {[x.as_posix() for x in self._incorrect_features]}"
            )
        if raise_if_nullable_id and self._nullable_id_features:
            raise FeaturesWithoutIDPresenceError(
                f"Features without IDs: {[x.as_posix() for x in self._nullable_id_features]}"
            )
        logger.info("All features are correct")



from typing import Optional

from overhave.base_settings import AuthorizationStrategy, OverhaveAuthorizationSettings
from overhave.entities import (
    OverhaveAdminSettings,
    OverhaveEmulationSettings,
    OverhaveFileSettings,
    OverhaveLanguageSettings,
    OverhaveLdapManagerSettings,
    OverhaveRedisSettings,
    OverhaveReportManagerSettings,
)
from overhave.factory.context.base_context import BaseFactoryContext
from overhave.scenario import OverhaveProjectSettings, OverhaveScenarioCompilerSettings
from overhave.test_execution import OverhaveTestSettings
from overhave.transport import OverhaveLdapClientSettings, OverhaveS3ManagerSettings


class OverhaveAdminContext(BaseFactoryContext):
    """Overhave admin context, based on application BaseSettings.

    This context defines how Overhave admin will work.
    """

    def __init__(
        self,
        admin_settings: Optional[OverhaveAdminSettings] = None,
        auth_settings: Optional[OverhaveAuthorizationSettings] = None,
        emulation_settings: Optional[OverhaveEmulationSettings] = None,
        file_settings: Optional[OverhaveFileSettings] = None,
        language_settings: Optional[OverhaveLanguageSettings] = None,
        ldap_manager_settings: Optional[OverhaveLdapManagerSettings] = None,
        ldap_client_settings: Optional[OverhaveLdapClientSettings] = None,
        redis_settings: Optional[OverhaveRedisSettings] = None,
        project_settings: Optional[OverhaveProjectSettings] = None,
        compilation_settings: Optional[OverhaveScenarioCompilerSettings] = None,
        report_manager_settings: Optional[OverhaveReportManagerSettings] = None,
        s3_manager_settings: Optional[OverhaveS3ManagerSettings] = None,
        test_settings: Optional[OverhaveTestSettings] = None,
    ) -> None:
        super().__init__(
            emulation_settings=emulation_settings or OverhaveEmulationSettings(),
            file_settings=file_settings or OverhaveFileSettings(),
            language_settings=language_settings or OverhaveLanguageSettings(),
            project_settings=project_settings or OverhaveProjectSettings(),
            compilation_settings=compilation_settings or OverhaveScenarioCompilerSettings(),
            report_manager_settings=report_manager_settings or OverhaveReportManagerSettings(),
            s3_manager_settings=s3_manager_settings or OverhaveS3ManagerSettings(),
            test_settings=test_settings or OverhaveTestSettings(),
        )
        self.admin_settings = admin_settings or OverhaveAdminSettings()
        self.auth_settings = auth_settings or OverhaveAuthorizationSettings()
        self.redis_settings = redis_settings or OverhaveRedisSettings()

        if self.auth_settings.auth_strategy is AuthorizationStrategy.LDAP:
            self.ldap_manager_settings = ldap_manager_settings or OverhaveLdapManagerSettings()
            self.ldap_client_settings = ldap_client_settings or OverhaveLdapClientSettings()



import abc
import logging
from uuid import uuid4

from pydantic import SecretStr

from overhave import db
from overhave.entities.auth_managers.base import BaseAdminAuthorizationManager
from overhave.storage import ISystemUserStorage

logger = logging.getLogger(__name__)
_ADMIN_USERNAME = "admin"


class AdminSecretMixin(BaseAdminAuthorizationManager, abc.ABC):
    """
    Special class for default admin user preparation.

    Usage: application will create 'admin' user with UUID randomly generated secret key.
    This user will be placed into database, moreover key will be logged with INFO logging level.
    Note: real service administrator should bring this key and change password or delete this user!
    """

    def __init__(self, system_user_storage: ISystemUserStorage):
        super().__init__(system_user_storage)
        self._make_secret()

    @staticmethod
    def _get_secret() -> SecretStr:
        return SecretStr(uuid4().hex)

    def _make_secret(self) -> None:
        user = self._system_user_storage.get_user_by_credits(login=_ADMIN_USERNAME)
        if user is not None:
            logger.info("Admin user already exists")
            return
        secret = self._get_secret()
        logger.info("Generated admin secret: %s", secret.get_secret_value())
        self._system_user_storage.create_user(login=_ADMIN_USERNAME, password=secret, role=db.Role.admin)



import abc
from functools import cached_property

from overhave.factory.base_factory import BaseOverhaveFactory, IOverhaveFactory
from overhave.factory.context import OverhaveSynchronizerContext
from overhave.scenario import FeatureValidator
from overhave.synchronization import IOverhaveSynchronizer, OverhaveSynchronizer, SynchronizerStorageManager


class ISynchronizerFactory(IOverhaveFactory[OverhaveSynchronizerContext]):
    """Factory interface for Overhave feature synchronizer application."""

    @property
    @abc.abstractmethod
    def synchronizer(self) -> IOverhaveSynchronizer:
        pass

    @property
    @abc.abstractmethod
    def feature_validator(self) -> FeatureValidator:
        pass


class SynchronizerFactory(BaseOverhaveFactory[OverhaveSynchronizerContext], ISynchronizerFactory):
    """Factory for Overhave feature synchronization application."""

    context_cls = OverhaveSynchronizerContext

    @cached_property
    def _synchronizer(self) -> OverhaveSynchronizer:
        return OverhaveSynchronizer(
            file_settings=self.context.file_settings,
            scenario_parser=self._scenario_parser,
            feature_extractor=self._feature_extractor,
            git_initializer=self._git_initializer,
            storage_manager=SynchronizerStorageManager(
                feature_storage=self._feature_storage,
                feature_type_storage=self._feature_type_storage,
                scenario_storage=self._scenario_storage,
                tag_storage=self._feature_tag_storage,
                draft_storage=self._draft_storage,
                system_user_storage=self._system_user_storage,
            ),
        )

    @property
    def synchronizer(self) -> IOverhaveSynchronizer:
        return self._synchronizer

    @cached_property
    def _feature_validator(self) -> FeatureValidator:
        return FeatureValidator(
            file_settings=self.context.file_settings,
            scenario_parser=self._scenario_parser,
            git_initializer=self._git_initializer,
        )

    @property
    def feature_validator(self) -> FeatureValidator:
        return self._feature_validator



import abc
from typing import List, Optional, cast

from overhave import db
from overhave.storage import TestUserModel, TestUserSpecification
from overhave.utils import get_current_time


class BaseTestUserStorageException(Exception):
    """Base exception for :class:`FeatureStorage`."""


class TestUserDoesNotExistError(BaseTestUserStorageException):
    """Error for situation when test user not found."""

    __test__ = False


class TestUserUpdatingNotAllowedError(BaseTestUserStorageException):
    """Error for situation when test user has allow_update=False."""

    __test__ = False


class ITestUserStorage(abc.ABC):
    """Abstract class for Test User storage."""

    @staticmethod
    @abc.abstractmethod
    def get_test_user_by_id(user_id: int) -> Optional[TestUserModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_test_user_by_name(name: str) -> Optional[TestUserModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_test_users(feature_type_id: int, allow_update: bool) -> List[TestUserModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def create_test_user(
        name: str,
        specification: TestUserSpecification,
        created_by: str,
        feature_type_id: int,
        allow_update: bool,
    ) -> TestUserModel:
        pass

    @staticmethod
    @abc.abstractmethod
    def update_test_user_specification(user_id: int, specification: TestUserSpecification) -> None:
        pass

    @staticmethod
    @abc.abstractmethod
    def delete_test_user(user_id: int) -> None:
        pass


class TestUserStorage(ITestUserStorage):
    """Class for Test User storage."""

    @staticmethod
    def get_test_user_by_id(user_id: int) -> Optional[TestUserModel]:
        with db.create_session() as session:
            user: Optional[db.TestUser] = session.query(db.TestUser).get(user_id)
            if user is not None:
                return cast(TestUserModel, TestUserModel.from_orm(user))
            return None

    @staticmethod
    def get_test_user_by_name(name: str) -> Optional[TestUserModel]:
        with db.create_session() as session:
            user: Optional[db.TestUser] = session.query(db.TestUser).filter(db.TestUser.name == name).one_or_none()
            if user is not None:
                return cast(TestUserModel, TestUserModel.from_orm(user))
            return None

    @staticmethod
    def get_test_users(feature_type_id: int, allow_update: bool) -> List[TestUserModel]:
        with db.create_session() as session:
            db_users: List[db.TestUser] = (
                session.query(db.TestUser)
                .filter(db.TestUser.feature_type_id == feature_type_id, db.TestUser.allow_update == allow_update)
                .all()
            )
            return [cast(TestUserModel, TestUserModel.from_orm(user)) for user in db_users]

    @staticmethod
    def create_test_user(
        name: str,
        specification: TestUserSpecification,
        created_by: str,
        feature_type_id: int,
        allow_update: bool,
    ) -> TestUserModel:
        with db.create_session() as session:
            test_user = db.TestUser(  # type: ignore
                name=name,
                specification=specification,
                feature_type_id=feature_type_id,
                created_by=created_by,
                allow_update=allow_update,
                changed_at=get_current_time(),
            )
            session.add(test_user)
            session.flush()
            return cast(TestUserModel, TestUserModel.from_orm(test_user))

    @staticmethod
    def update_test_user_specification(user_id: int, specification: TestUserSpecification) -> None:
        with db.create_session() as session:
            test_user = session.query(db.TestUser).get(user_id)
            if test_user is None:
                raise TestUserDoesNotExistError(f"Test user with id {user_id} does not exist!")
            if not test_user.allow_update:
                raise TestUserUpdatingNotAllowedError(f"Test user updating with id {user_id} not allowed!")
            test_user.specification = specification
            test_user.changed_at = get_current_time()

    @staticmethod
    def delete_test_user(user_id: int) -> None:
        with db.create_session() as session:
            user: Optional[db.TestUser] = session.query(db.TestUser).get(user_id)
            if user is None:
                raise TestUserDoesNotExistError(f"Test user with id {user_id} does not exist!")
            session.delete(user)



import logging
from typing import Optional

from _pytest.main import Session

from overhave.entities import IFeatureExtractor, OverhaveFileSettings, StepPrefixesModel
from overhave.test_execution import PytestRunner, StepCollector

logger = logging.getLogger(__name__)


class BaseConfigInjectorException(Exception):
    """Base exception for :class:`ConfigInjector`."""


class NullableEntitiesError(BaseConfigInjectorException):
    """Excepion for situation with unresolved entities."""


class ConfigInjector:
    """Special class to operate with pytest entites, for example get data and patch objects."""

    def __init__(self) -> None:
        self._file_settings: Optional[OverhaveFileSettings] = None
        self._step_collector: Optional[StepCollector] = None
        self._test_runner: Optional[PytestRunner] = None
        self._feature_extractor: Optional[IFeatureExtractor] = None

        self._current_type: Optional[str] = None
        self._initialized = False

    @staticmethod
    def patch_pytestbdd_prefixes(custom_step_prefixes: Optional[StepPrefixesModel]) -> None:
        """Def for patch pytestbdd STEP_PREFIXES."""
        if custom_step_prefixes is not None:
            step_prefixes = custom_step_prefixes.extend_defaults()
            logger.debug("Successfully extended pytest-bdd step prefixes with custom prefixes:\n%s", step_prefixes)

    def supplement_components(
        self,
        file_settings: OverhaveFileSettings,
        step_collector: StepCollector,
        test_runner: PytestRunner,
        feature_extractor: IFeatureExtractor,
    ) -> None:
        """Def for dynamical supplement of settings on-the-fly (strictly after PyTest configuring)."""
        self._file_settings = file_settings
        self._step_collector = step_collector
        self._test_runner = test_runner
        self._feature_extractor = feature_extractor

    def collect_configs(self) -> None:
        """Special def for pytest run in collect-only mode. Collects pytestbdd steps and other config data."""
        if self._feature_extractor is None or self._test_runner is None or self._file_settings is None:
            raise NullableEntitiesError("Could not collect configuration for injection!")
        if not self._initialized:
            logger.info("Started initialization process...")
            for feature_type in self._feature_extractor.feature_types:
                self._current_type = feature_type
                self._test_runner.collect_only(
                    fixture_file=(
                        self._file_settings.fixtures_dir
                        / self._file_settings.fixtures_file_template_mask.format(feature_type=feature_type)
                    ),
                )
            self._current_type = None
            logger.info("Successfully initialized.")
            self._initialized = True

    @property
    def _inside_collecting(self) -> bool:
        return self._current_type is not None and not self._initialized

    def _load_from_session(self, session: Session) -> None:
        """Loads pytestbdd steps and other config data from pytest Session."""
        if self._step_collector is None:
            raise NullableEntitiesError
        if self._inside_collecting:
            self._step_collector.collect_steps(session=session, feature_type=self._current_type)  # type: ignore

    def adapt(self, session: Session) -> None:
        """Main def for data loading from pytest Session and enrichment of pytest configs."""
        if self._inside_collecting:
            self._load_from_session(session)



from typing import Any, Optional

import allure
from _pytest.nodes import Item

from overhave.pytest_plugin.helpers.parsed_info import get_feature_info_from_item
from overhave.scenario import OverhaveScenarioCompilerSettings


def _get_severity_level_from_tags(item: Item, keyword: str) -> Optional[allure.severity_level]:
    for marker in reversed(item.own_markers):
        if not marker.name.startswith(keyword):
            continue
        severity = marker.name.removeprefix(keyword)
        return allure.severity_level(severity)
    return None


def _get_parsed_feature_severity(item: Item, **kwargs: Any) -> Optional[allure.severity_level]:
    return get_feature_info_from_item(item).severity


def _get_default_severity(*args: Any, **kwargs: Any) -> allure.severity_level:
    return allure.severity_level.NORMAL


def set_severity_level(
    compilation_settings: OverhaveScenarioCompilerSettings,
    item: Item,
) -> None:
    for extractor_func in _get_severity_level_from_tags, _get_parsed_feature_severity, _get_default_severity:
        severity_lvl = extractor_func(item=item, keyword=compilation_settings.severity_keyword)  # type: ignore
        if severity_lvl is None:
            continue
        allure.dynamic.severity(severity_lvl)
        return



import logging
from typing import Optional

from pydantic import SecretStr

from overhave.entities.auth_managers.secret_mixin import AdminSecretMixin
from overhave.storage import SystemUserModel

logger = logging.getLogger(__name__)


class BaseSimpleAdminAuthorizationManagerException(Exception):
    """Base exception for :class:`SimpleAdminAuthorizationManager`."""


class NullablePasswordError(Exception):
    """Exception for situation when row is without password."""


class SimpleAdminAuthorizationManager(AdminSecretMixin):
    """Class for user registration.

    Manager does not provide real auth_managers. Each user could use preferred name.
    This name will be used for user authority. Each user is unique. Passwords not required.
    """

    def authorize_user(self, username: str, password: SecretStr) -> Optional[SystemUserModel]:
        user = self._system_user_storage.get_user_by_credits(login=username)
        if user is None:
            user = self._system_user_storage.create_user(login=username, password=password)
        if user.password is None:
            raise NullablePasswordError(f"User with id={user.id} has not got password!")
        if user.password.get_secret_value() == password.get_secret_value():
            return user
        return None
